# pygame_ext.py -- extensions to Pygame for angrydd
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: pygame_ext.py 286 2004-09-04 03:51:59Z piman $"

import os
import util
from pygame import mixer, error, time

# Override the standard mixer_music methods with ones that can
# fail gracefully if the mixer can't start or the music file
# is missing.
def new_load(filename):
    try: mixer.music._load(filename)
    except error:
        print(("W: Unable to load %s " % filename), end=''),
        if os.path.exists(filename): print("(is your soundcard in use?)")
        else: print("(where's the full sounds/ directory?)")
        pass

def new_play(loop):
    try: mixer.music._play(loop)
    except error: pass

def new_fadeout(time):
    try: mixer.music._fadeout(time)
    except error: pass

mixer.music._load = mixer.music.load
mixer.music._play = mixer.music.play
mixer.music._fadeout = mixer.music.fadeout
mixer.music.load = new_load
mixer.music.play = new_play
mixer.music.fadeout = new_fadeout

# Override time.get_ticks() to subtract an offset; this allows simple
# pausing of game for all objects using get_ticks().
time._get_ticks = time.get_ticks
time.paused = False
time.__start_pause = 0
time.__at_pause = 0
time.__offset = 0

time.clock = util.WaitClock()

def pause():
    if not time.paused:
        time.__start_pause = time._get_ticks()
        time.__at_pause = time.get_ticks()
        time.paused = True
time.pause = pause

def unpause():
    if time.paused:
        time.__offset += time._get_ticks() - time.__start_pause
        time.paused = False
time.unpause = unpause

def toggle_pause():
    if time.paused: time.unpause()
    else: time.pause()
time.toggle_pause = toggle_pause

def new_get_ticks():
    if time.paused: return time.__at_pause
    else: return time._get_ticks() - time.__offset
time.get_ticks = new_get_ticks

# Override Sound. This returns a dummy object when the sound file could
# not be loaded (busy soundcard, missing file, bad SDL_mixer...)
mixer._Sound = mixer.Sound
def new_Sound(filename):
    try: return mixer._Sound(filename)
    except error:
        print(("W: Unable to load %s " % filename), end='')
        return FakeSound()

mixer.Sound = new_Sound

# Pretend to be a sound. FIXME: There might be some way to check
# function arity and detect more calling problems ahead of time.
class FakeSound(object):
    def __getattr__(self, name):
        if not isinstance(mixer._Sound, type) or hasattr(mixer._Sound, name):
            if name.startswith("get_"): return (lambda *args: 0)
            else: return (lambda *args: None)
        else: raise AttributeError(name)



# howtoplay.py -- a basic help screen for angrydd
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: howtoplay.py 286 2004-09-04 03:51:59Z piman $"

import pygame; from pygame import transform
import characters; from characters import Character

import config
import textfx
import wipes
import load

from events import EventManager
from boxes import SpecialSprite

from constants import *

# Display the "how to play" screen (with appropriate unlocks enabled).
# FIXME: This is pretty messy; the text belongs in a data file rather
# than inline.
def init(*args):
    em = EventManager()
    screen = pygame.display.get_surface()
    move_snd = load.sound("select-move.wav")

    index = 0

    # The format is up to three paragraphs per screen, displayed
    # left, right, left,
    texts = [
        ["In Angry, Drunken Dwarves, you are an angry, drunken dwarf. Why "
         "are you so angry? Who knows. But you've decided to take your "
         "aggression out on other dwarves, by dropping gems on their "
         "heads.",
         "Multicolored gems will fall from the top of the screen. You "
         "should try to arrange them into groups by color. The arrow keys "
         "or a/d and j/l will move them left and right. s or k will make "
         "then fall faster, and q/e and u/o will rotate them.",
         "As gems reach the bottom, they will land and sit in place. If you "
         "put enough gems of the same color in a rectangle shape, they will "
         "merge and form a crystal."
         ],
        ["Less common are break gems. When you drop a break gem on a gem "
         "or crystal of the same color, it will destroy any gems of that "
         "color adjacent to it. Crystals are worth much more than an equal "
         "number of regular gems.",
         "When you destroy gems on your side, it will drop counter gems "
         "onto your opponent. These are harder to destroy, but turn into "
         "regular gems after a few turns. You can try to cancel an incoming "
         "attack by breaking gems yourself, but you'll need to break twice "
         "as many as are coming.",
         "Rarer is the diamond. When you drop a diamond onto something, "
         "it will destroy all gems of that color on your side, even "
         "counter gems! Be careful though, because a crystal "
         "destroyed by a diamond isn't worth any more than normal gems."
         ],
        ["Each dwarf has a drop gem pattern, and an attack strength. The drop "
         "pattern is a display of what kinds of gems will be dropped when "
         "you attack your opponent. Small attacks will drop the same "
         "row over and over, which will probably help them!",
         "The attack strength is a measurement of how strong your drop "
         "pattern is; drops with weirder patterns are harder to plan "
         "against. Before gems are dropped, they are multiplied by this "
         "attack strength, and so the damage is scaled up or down.",
         "For the most devestating results, try chaining attacks together, "
         "so breaking some gems results in even more breaking afterwards. "
         "Breaking many crystals in a chain can result in huge drops."
         ]
        ]

    drop = pygame.Surface([64, 64])
    drop.blit(load.block("blue"), [0, 0])
    drop.blit(load.block("yellow"), [0, 32])
    drop.blit(load.block("green"), [32, 0])
    drop.blit(load.block("red"), [32, 32])

    crystal = load.gem("green", 4, 3)

    breaks = pygame.Surface([64, 64])
    breaks.blit(load.block("red", "-crash"), [0, 0])
    breaks.blit(load.block("green", "-crash"), [0, 32])
    breaks.blit(load.block("yellow", "-crash"), [32, 0])
    breaks.blit(load.block("blue", "-crash"), [32, 32])

    counters = pygame.Surface([64, 64])
    counters.blit(textfx.lettered_box("5", "green"), [0, 0])
    counters.blit(textfx.lettered_box("4", "blue"), [32, 0])
    counters.blit(textfx.lettered_box("3", "red"), [0, 32])
    counters.blit(textfx.lettered_box("2", "yellow"), [32, 32])

    chain = pygame.Surface([64, 96])
    chain.blit(load.block("blue", "-crash"), [0, 64])
    chain.blit(load.block("green"), [32, 64])
    chain.blit(load.block("green", "-crash"), [32, 32])
    chain.blit(load.block("blue"), [32, 0])

    # Followed by up to three images per screen, right, left, right.
    images = [
        [None, drop, crystal],
        [breaks, counters, load.block("diamond")],
        [Character.arcade[4].drop.render(), None, chain],
        ]

    if config.getboolean("unlock", "single"):
        texts.append([
            "In single player mode, rather than competing against someone "
            "else, you're racing the clock. You have to clear a certain "
            "number of blocks in a certain number of turns, or the ones "
            "left get dumped on you.",
            "Your field is also twice as big. So it sounds easy, right? "
            "Well, to start with, you've got three new colors of gems to "
            "contend with: orange, purple, and cyan.",
            "If that wasn't bad enough, the number of gems you have to clear "
            "goes up much faster than the number of turns you have "
            "to do it, so build up those crystals early and save them."
            ])

        newcols = pygame.Surface([64, 64])
        newcols.blit(load.block("orange"), [16, 0])
        newcols.blit(load.block("cyan"), [0, 32])
        newcols.blit(load.block("purple"), [32, 32])

        
        counts = pygame.Surface([130, 100])
        box = Character.default.border([40, 40])
        text1 = textfx.shadow("new:", 20)
        text2 = textfx.shadow("turns:", 20)
        num1 = textfx.shadow("122", 30)
        num2 = textfx.shadow("10", 30)
        counts.blit(box, [0, 20])
        counts.blit(box, [70, 20])
        counts.blit(text1, text1.get_rect(center = [30, 38]))
        counts.blit(text2, text2.get_rect(center = [100, 38]))
        counts.blit(num1, [12, 45])
        counts.blit(num2, [87, 45])
        images.append([None, newcols, counts])

    if config.getboolean("unlock", "combat"):
        texts.append([
            "Combat blocks look like normal gems with a special symbol "
            "in the middle. These gems don't form crystals, but otherwise "
            "break like normal colored gems. When you break one of these, "
            "you 'pick up' the special attack in it.",
            "To use the attack, press start (enter/2). There are five basic "
            "attacks; from left to right: make your opponent's field blink, "
            "clear  your own field of all blocks,",
            "flip your opponent's screen upside down, disable the 'next' "
            "indicator, drop some gray blocks, or reverse your opponent's "
            "controls. Blink, flip, and reverse last for a few "
            "seconds. Scramble lasts 10 turns."])

        blink = load.block("blue")
        clear = load.block("red")
        rev = load.block("yellow")
        flip = load.block("green")
        gray = load.block("red")
        scram = load.block("purple")
        blink.blit(SpecialSprite.load(BLINK), [6, 6])
        flip.blit(SpecialSprite.load(FLIP), [6, 6])
        clear.blit(SpecialSprite.load(CLEAR), [6, 6])
        rev.blit(SpecialSprite.load(REVERSE), [6, 6])
        gray.blit(SpecialSprite.load(GRAY), [6, 6])
        scram.blit(SpecialSprite.load(SCRAMBLE), [6, 6])
        img1 = pygame.Surface([64, 32])
        img2 = pygame.Surface([64, 64])
        img1.blit(blink, [0, 0])
        img1.blit(clear, [32, 0])
        img2.blit(flip, [0, 0])
        img2.blit(scram, [32, 0])
        img2.blit(gray, [0, 32])
        img2.blit(rev, [32, 32])

        images.append([None, img1, img2])

    wipes.wipe_in(render_help_page(texts[index], images[index]))

    cont = True
    screen.blit(render_help_page(texts[index], images[index]), [0, 0])
    img = textfx.shadow("Enter: Menu - Left/Right: Turn Page (%d/%d)" %
                 (index + 1, len(images)), 18)
    screen.blit(img, [785 - img.get_width(), 10])
    pygame.display.update()
    while cont:
        oldindex = index
        for ev in em.wait():
            if ev.type == QUIT: cont = False
            elif ev.type == PLAYER:
                if ev.key == CONFIRM: cont = False
                elif ev.key == UP or ev.key == LEFT:
                    index = (index - 1) % len(texts)
                    move_snd.play()
                elif ev.key in [DOWN, RIGHT, ROT_CC, ROT_CW]:
                    index = (index + 1) % len(texts)
                    move_snd.play()

        if oldindex != index:
            screen.blit(render_help_page(texts[index], images[index]), [0, 0])
            img = textfx.shadow("Enter: Menu - Left/Right: Turn Page (%d/%d)" %
                         (index + 1, len(images)), 18)
            screen.blit(img, [785 - img.get_width(), 10])
            pygame.display.update()

    return True

# Render a page in the format used above.
def render_help_page(texts, images):
    surf = Character.default.border([780, 580])
    font = textfx.WrapFont(32, 450)

    for i in range(len(texts)):
        text = font.render(texts[i])
        image = images[i]
        if image:
            ri = image.get_rect()
            image.set_colorkey(image.get_at([0, 0]))
        rt = text.get_rect()
        if i & 1:
            if image: ri.centerx = 200
            rt.centerx = 500
        else:
            rt.centerx = 300
            if image: ri.centerx = 630

        if image: ri.centery = 95 + (200 * i)
        rt.centery = 95 + (200 * i)
        surf.blit(text, rt)
        if image: surf.blit(image, ri)
    return surf



# util.py -- utility fuctions/classes for angrydd
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: util.py 286 2004-09-04 03:51:59Z piman $"

import pygame

# Implement the singleton pattern.
# In reality this didn't work like I thought, because in Python
# __new__ doesn't call __init__.
class Singleton(object):
    def __new__(*args):
        if not hasattr(args[0], "_singleton"):
            return super(Singleton, args[0]).__new__(*args)
        else: return args[0]._singleton

    def __init__(self):
        type(self)._singleton = self

# "Flatten" a list
def flatten(sequence, is_scalar = lambda x: not isinstance(x, list)):
    result = []
    for item in sequence:
        if is_scalar(item): result.append(item)
        else: result.extend(flatten(item))
    return result

# Regular pygame.time.Clock uses pygame.time.delay; this uses time.wait.
# That way, we save CPU time on Unix.
class WaitClock(object):
    def __init__(self, time_func = pygame.time.get_ticks):
        self._time = time_func
        self._last_tick = 0

    def tick(self, fps = 60):
        t = self._time()
        sleep = (1000 // fps) - (t - self._last_tick)
        self._last_tick = t
        if sleep > 0: pygame.time.wait(sleep)



# dirstore - generic access to archive formats
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""This module provides a generic object (DirStore) to access files
stored in a directory, tar, rar, or zip archive. You can read
data from the archive as a string or a file-like object without
worrying about the underlying archive format.

It also provides a MetaStore object which contains multiple
DirStores, but acts as a single one. When a name is requested
from a MetaStore it returns the first object it finds in its
stores with that name.

It supports the following formats:
 (This information is inaccurate if dirstore is not currently in use)
"""

import os
import stat
import fnmatch

try: from io import StringIO
except ImportError: import StringIO

# Trick borrowed from Peter Norvig's Python IAQ:
# http://www.norvig.com/python-iaq.html
def _abstract():
    import inspect
    caller = inspect.getouterframes(inspect.currentframe())[1][3]
    raise NotImplementedError(caller + ' must be implemented in subclass')

# Defines available formats. The first is a predicate function that
# returns True if the second, a constructor, should be used.
FORMATS = []

# TODO:
# - A function that extracts a file and writes it to somewhere else.

version = '0.1.0'
__author__ = 'Joe Wreschnig <piman@sacredchao.net>'
__revision__ = "$Id: dirstore.py 32 2004-05-29 06:07:17Z piman $"
__license__ = "GNU GPL v2"

def require(ver):
    """You can use this function to require a specific version of the
dirstore module, e.g. dirstore.require("0.1.0") requires at least version
0.1.0. You don't need to specify trailing zeros, so the above example is
equivalent to dirstore.require("0.1").

If the dirstore module is not recent enough, an ImportError will be
raised."""
    if ver.split(".") > version.split("."):
        raise ImportError("you need at least dirstore %s to run this (you have %s)" % (ver, version))

class DirStoreError(RuntimeError):
    """All exceptions specific to DirStores will be subclasses of this
type of exception."""
    pass

class UnknownStoreType(DirStoreError):
    """This exception is raised when you try to open an unsupported file
type as a DirStore."""

    def __init__(self, filename):
        DirStoreError.__init__(self)
        self.__filename = filename

    def __str__(self):
        return "%s is of an unknown/unsupported file type" % self.__filename

class WrongStoreError(DirStoreError):
    """This exception is raised when you pass a FileInfo object to
a DirStore, and that FileInfo did not originally come from that
DirStore."""

class InvalidStore(DirStoreError):
    """This exception is raised when something in a DirStore cannot be
opened properly. It differs from UnknownStoreType, in that if this
exception is raised, the file's type was (possibly incorrectly)
detected, but could not actually be read."""
    pass

class FileInfo(object):
    """FileInfo objects are the return type of the [] and info methods
of DirStores. They contain two important member, name and size.
name is the name of the object within whatever store it came from;
size is the (uncompressed) size of the object in bytes. If the size
could not be determined, this is -1."""

    def __init__(self, **kwargs):
        self.__dict__.update({ "size": -1 })
        self.__dict__.update(kwargs)

    def open(self):
        """Return a file-like object containing the contents
of this entry in the store.

store.info(name).open() is equivalent to store.open(name)."""
        return self._store.open(self.name)

    def read(self):
        """Return a string containing the contents
of this entry in the store.

store.info(name).read() is equivalent to store.read(name)."""
        return self._store.read(self.name)

class MetaStore(object):
    """A metastore is an object that acts like a single store, but is
actually a store of stores. It supports all the methods of a
DirStore, which apply to the first store with the given name found
in it.

It also supports additional methods for accessing the stores in it.
In particular, it supports the [], pop, remove, insert, append, and
extend operations just like a list does. It also supports slicing.

Note that iterating through a MetaStore returns the members of all of
its stores, not the stores themselves.

MetaStores do not have ZipFile compatibility methods."""
    def __init__(self, stores = []):
        """Initialize a metastore. The list of stores can a string (valid
for the DirStore constructor), or any DirStore-like object."""

        # Convert strings to stores.
        self.__stores = [(((str(s) == s) and DirStore(s)) or s)
                         for s in stores]

    def __iter__(self): return iter([self.info(name) for name in self.ls()])
    def append(self, store):  self.__stores.append(store)
    def extend(self, stores): self.__stores.extend(stores)
    def __getitem__(self, i): return self.__stores[i]
    def __delitem__(self, i): del(self.__stores[i])
    def __setitem__(self, i, store): self.__stores[i] = store
    def __delslice__(self, i, j): del(self.__stores[i:j])
    def __getslice__(self, i, j): return self.__stores[i:j]
    def __setslice__(self, i, j, y): self.__stores[i:j] = y
    def pop(self, index = -1): return self.__stores.pop(index)
    def remove(self, value): self.__stores.remove(value)
    def insert(self, i, store): self.__stores.insert(i, store)

    def verify(self):
        ret = DirStore.OK
        for store in self.__stores:
            v = store.verify()
            if v == DirStore.BAD: return DirStore.BAD
            elif v == DirStore.UNKNOWN: ret = DirStore.UNKNOWN
        return ret

    def ls(self, glob = None):
        names = []
        for store in self.__stores: names.extend(store.ls(glob))
        # only include each name once
        names = list(dict(map(None, names, [])).keys())
        names.sort()
        return names

    def read(self, name):
        return self.find(name).read(name)

    def open(self, name):
        return self.find(name).open(name)

    def info(self, name):
        return self.find(name).info(name)

    def info_all(self, name):
        """Returns FileInfo objects for every object with the given
name, in every store."""
        return [s.info(name) for s in self.find_all(name)]

    def find(self, name):
        """Find the first store with the given name in it."""
        for store in self.__stores:
            if name in store: return store
        else: raise ValueError("%s not found in any contained store" % name)

    def find_all(self, name):
        """Like find, but returns all stores with the given name in them."""
        stores = []
        for store in self.__stores:
            if name in store: stores.append(store)
        return stores

# This store is an abstract type. It should be subclassed by types
# that implement its abstract functions.
#
# Certain functions are given some default implementations. These are
# defined in terms of other functions that must be implemented, and
# are usually not the most efficient way to do it. So, you should
# override them if at all possible.
class DirStore(object):
    """This is the basic store type. Certain other types of stores
subclass this, and when you instantiate this, you will actually get
one of its subclasses. However, you shouldn't instantiate the
subclasses directly.

Aliases exist to provide compatibility with ZipFile and TarFileCompat:
  getinfo  -> info
  namelist -> ls
  testzip  -> verify
  read     -> read

close, infolist, printdir, write, and writestr have no equivalents.
 """

    # Return codes for DirStore.verify.
    BAD, OK, UNKNOWN = range(3)

    def __new__(cls, filename):
        """Use this to create a DirStore, e.g. 'DirStore("foo/bar.zip").
This will return an instance of one of DirStore's child classes;
which exact one doesn't matter. You shouldn't instantiate these
classes directly, since their names might change in the future,
or the way in which they are instantiated. Using
isinstance(some_store, DirStore) will always be true, however.

All stores have one public member, self.name, which is the name
the store was created with.

All the functions in a store that can take a name, can also take
a FileInfo object from that store.

If you request a name that is not in the store, a ValueError will
be raised."""

        if (cls == DirStore):
            for pred, Ctr in FORMATS:
                if pred(filename): return Ctr(filename)
            raise UnknownStoreType(filename)
        else:
            return super(DirStore, cls).__new__(cls)
    
    def __init__(self, filename):
        self.name = os.path.abspath(filename)
        self.getinfo = self.info
        self.namelist = self.ls
        self.testzip = self.verify

    def _read(self, name): _abstract()
    def _info(self, name): _abstract()
    def _ls(self): _abstract()

    def ls(self, glob = None):
        """Return a list of names available within the store. Any
names in 'subdirectories' will be returned as something
like 'dir/name'.

The optional glob argument is a string to filter results on (via
Python's fnmatch module)."""
        if glob is None: return self._ls()
        else: return fnmatch.filter(self._ls(), glob)

    def info(self, name):
        """Return a FileInfo object describing the member of the store
with the given name."""
        if name not in self:
            raise ValueError("%s is not in this store" % name)
        try:
            if name._store is self: return self._info(name.name)
            else: raise WrongStoreError("%s is not from %s" % (name, self))
        except AttributeError: return self._info(name)

    def read(self, name):
        """Read the data in the member with the given name and return
it as a string."""
        if name not in self:
            raise ValueError("%s is not in this store" % name)
        try:
            if name._store is self: return self._read(name.name)
            else: raise WrongStoreError("%s is not from %s" % (name, self))
        except AttributeError: return self._read(name)

    def verify(self):
        """Verify the integrity of the archive. Returns zero when
the archive is known to be bad; non-zero otherwise. Possible
return values are currently DirStore.BAD, DirStore.OK, and
DirStore.UNKNOWN."""
        return DirStore.UNKNOWN

    def __iter__(self):
        """DirStores are iteraterable objects. The member names are
given in succession, as FileInfo objects."""
        return iter([self.info(name) for name in self.ls()])

    def __contains__(self, name):
        try:
            if name._store is self: return (name.name in self.ls())
            else: return False
        except AttributeError:
            return (name in self.ls())

    def __getitem__(self, name):
        """Equivalent to DirStore.info, you can access members by
store[name], giving a FileInfo object."""
        return self.info(name)

    def open(self, name):
        """Return a file-like object containing the data in the
given member."""
        if name not in self:
            raise ValueError("%s is not in this store" % name)
        try:
            if name._store is self: return self._open(name.name)
            else: raise WrongStoreError("%s is not from %s" % (name, self))
        except AttributeError: return self._open(name)

    def _open(self, name):
        return StringIO(self.read(name))

def _visit(arg, dirname, names):
    ns, selfname = arg
    for name in names:
        fname = os.path.join(dirname, name)
        if os.path.isfile(fname):
            ns.append(fname[len(selfname) + 1:])

class _FSStore(DirStore):
    def verify(self):
        if os.path.isdir(self.name): return DirStore.OK
        else: return DirStore.BAD

    def _ls(self):
        try:
            return self._namelist
        except AttributeError:
            names = []
            arg = (names, self.name)
            for (dirpath, dirnames, filenames) in os.walk(self.name):
                _visit(arg, dirpath, filenames + dirnames)
            self._namelist = names
            return self._namelist

    def _open(self, name):
        name = os.path.join(self.name, name)
        return open(name, "rb")

    def _info(self, name):
        fullname = os.path.join(self.name, name)
        st = os.stat(fullname)
        return FileInfo(name = name, size = st[stat.ST_SIZE], _store = self)

    def _read(self, name):
        name = os.path.join(self.name, name)
        return open(name).read()
    
__doc__ += " - Directories in a local filesystem hierarchy. (Enabled)\n"
FORMATS.append((os.path.isdir, _FSStore))

try:
    __doc__ += " - ZIP compressed archives. "
    import zipfile

    class _ZipStore(DirStore):
        def __init__(self, filename):
            DirStore.__init__(self, filename)
            self.__zip = zipfile.ZipFile(filename)

        def verify(self):
            if self.__zip.testzip() is None: return DirStore.OK
            else: return DirStore.BAD

        def _info(self, name):
            inf = self.__zip.getinfo(name)
            return FileInfo(name = inf.filename, size = inf.file_size,
                            _store = self)

        def _ls(self): return self.__zip.namelist()

        def _read(self, name):
            return self.__zip.read(name)

    __doc__ += "(Enabled)\n"
    FORMATS.append(((lambda x: os.path.isfile(x) and zipfile.is_zipfile(x)),
                    _ZipStore))
except ImportError: __doc__ += "(Disabled)\n"

try:
    __doc__ += " - Tar compressed archives (gzip or bzip2). "
    import tarfile

    class _TarStore(DirStore):
        def __init__(self, filename):
            DirStore.__init__(self, filename)
            try:
                self.__tar = tarfile.TarFile.open(filename)
            except tarfile.ReadError:
                raise InvalidStore("%s could not be read" % filename)
            except tarfile.CompressionError:
                raise InvalidStore("%s could not be decompressed" % filename)

        def _ls(self): return self.__tar.getnames()

        def _info(self, name):
            return FileInfo(name = name, _store = self)

        def _open(self, name):
            return self.__tar.extractfile(name)

        def _read(self, name):
            return self.__tar.extractfile(name).read()

    __doc__ += "(Enabled)\n"
    FORMATS.append(((lambda x: os.path.isfile(x) and tarfile.is_tarfile(x)),
                    _TarStore))
except ImportError: __doc__ += "(Disabled)\n"

try:
    __doc__ += " - RAR archives. "

    if os.name != "posix": raise ImportError

    for path in os.environ["PATH"].split(os.pathsep):
         if os.path.isfile(os.path.join(path, 'unrar')):
             break
    else: raise ImportError

    class _RarStore(DirStore):
        def __init__(self, filename):
            DirStore.__init__(self, filename)
            fls = os.popen2(["unrar", "vb", "-c-", filename])[1].readlines()
            self.__files = [name.strip() for name in fls]

        def _ls(self): return list(self.__files)
        def _info(self, name):
            return FileInfo(name = name, _store = self)

        def _read(self, name):
            f = os.popen2(["unrar", "p", "-c-", "-inul",
                           self.name, name])[1]
            return f.read()

        def _open(self, name):
            return os.popen2(["unrar", "p", "-c-", "-inul",
                              self.name, name])[1]

    __doc__ += """(Enabled)
    Note that RAR support comes via your unrar binary and popen,
    and so can be very fragile. The RAR format is not recommended."""
    FORMATS.append((lambda s:os.path.isfile(s) and s.lower().endswith(".rar"),
                    _RarStore))
except ImportError: __doc__ += "(Disabled)\n"



# constants.py -- constants for angrydd
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: constants.py 295 2005-08-13 07:22:00Z piman $"

# This file is designed to import via 'from constants import *'.
# Accordingly, this really should only contain constants and nothing else!

import os
from pygame.locals import *

angrydd_path = os.path.split(os.path.realpath(__file__))[0]

# Using 'isdir' here causes problems on filesystems that can read
# .zip files as directories.
if angrydd_path.lower().endswith(".zip"):
    angrydd_path = os.path.split(angrydd_path)[0]

if os.name == "posix" and "HOME" in os.environ:
    rc_file = os.environ["HOME"] + "/.angryddrc"
else: rc_file = os.path.join(angrydd_path, "angryddrc")

PLAYER, PAUSE = range(USEREVENT + 1, USEREVENT + 3)

# Each of these must have a corresponding file in images/<color>.png,
# images/<color>-crash.png.
COLORS = ["blue", "green", "red", "yellow", "purple", "cyan", "orange"]

# Translate event keys.
UP, LEFT, RIGHT, DOWN, ROT_CW, ROT_CC, CONFIRM = range(7)

# Combat specials.
HEALTHY, CLEAR, REVERSE, FLIP, BLINK, GRAY, SCRAMBLE = range(7)

# AI difficulties
STUPID, VEASY, EASY, NORMAL, HARD, INSANE = range(6)

CREDITS = [
    "Angry, Drunken Dwarves",
    "Code & Design by Joe Wreschnig",
    "Music by theGREENzebra and JZig",
    "Characters by Zach Welhouse",
    "Portraits by Jessi Silver",
    "Additional Art by Jenn Hartnoll",
    "Voices by Chris Thorp",
    "Thanks to testers/helpers:",
    "antont, "
    "asterick, "
    "CrashChaos, ",
    "drewp, "
    "illume, "
    "jacius, "
    "Knio, ",
    "mindy, "
    "mojo, "
    "mu, "
    "Safiiru, ",
    "ShredWheat, "
    "sku, "
    "stargazer, ",
    "wm_eddie, "
    "yossarian, ",
    "people at MAS, "
    "and more I forgot."
    ]



# menu.py -- display/use the game and configuration menu
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: menu.py 306 2005-10-06 19:46:26Z piman $"

import pygame
from pygame.sprite import Sprite, RenderUpdates

import config
import textfx
import load

from util import Singleton
import events; from events import EventManager

from constants import *

# One of the carts displayed in the game menus.
class Platform(Sprite):
    def __init__(self, midbottom, string, pos):
        Sprite.__init__(self)
        self._pos = pos
        self.rect = Rect([0, 0, 0, 0])
        self.rect.midbottom = midbottom
        self.text = string
        self._lastx = self._goalx = self.rect.left
        self._ntime = 0

    # Set the text on the cart, using lettered box.es
    def _set_text(self, string):
        midbottom = self.rect.midbottom
        plat = load.image("platform.png")
        plat.set_colorkey(plat.get_at([0, 0]), RLEACCEL)
        plat_r = plat.get_rect()
        letters = textfx.boxes_from_string(string, pos = self._pos)
        surf = pygame.Surface([max(260, letters.get_width()),
                                   plat.get_height() + letters.get_height()])
        plat_r.centerx = surf.get_rect().centerx
        plat_r.bottom = surf.get_height() - 1
        surf.blit(plat, plat_r)
        letters_r = letters.get_rect()
        letters_r.centerx = surf.get_rect().centerx
        letters_r.bottom = surf.get_height() - 16
        surf.blit(letters, letters_r)
        surf.set_colorkey(surf.get_at([0, 0]), RLEACCEL)
        self.image = surf
        self.rect = self.image.get_rect()
        self.rect.midbottom = midbottom

    text = property(None, _set_text)

    # Move the cart to the left.
    def left(self):
        t = pygame.time.get_ticks()
        self._ntime = t + 600
        self._goalx += 260
        self._lastx = self.rect.left

    # Move it to the right.
    def right(self):
        t = pygame.time.get_ticks()
        self._ntime = t + 600
        self._goalx -= 260
        self._lastx = self.rect.left

    def update(self, time):
        if time < self._ntime:
            p = (self._ntime - time) / 600.0
            self.rect.left = self._goalx - (self._goalx - self._lastx) * p
        else: self.rect.left = self._goalx

# Top score display
class TopScores(Sprite, Singleton):
    def __init__(self):
        Sprite.__init__(self)
        self._start = pygame.time.get_ticks()
        self.read_scores()
        Singleton.__init__(self)

    # Load all the scores from the configuration data.
    def read_scores(self):
        self._images = []
        secs = []
        for c in range(1, 10, 2):
            sec = "scores-versus-%d" % c
            name = "Top Scores: Best %d/%d" % (c // 2 + 1, c)
            secs.append((sec, name))
        if config.getboolean("unlock", "single"):
            secs.append(("scores-single", "Top Scores: Single Player"))
        secs.append(("scores-versus-arcade", "Top Scores: Arcade"))

        for sec, name in secs:
            img = pygame.Surface([420, 110], SRCALPHA, 32)
            img.fill([0, 0, 0, 0])
            t = textfx.shadow(name, 26)
            img.blit(t, t.get_rect(midtop = [210, 0]))
            for i in range(1, 4):
                parts = config.get(sec, str(i)).split(",")
                score = parts.pop()
                name = ",".join(parts)
                tl = textfx.shadow(name, 30)
                tr = textfx.shadow(score, 30)
                img.blit(tl, tl.get_rect(topleft = [10, 25 * i]))
                img.blit(tr, tr.get_rect(topright = [415, 25 * i]))
            self._images.append(img)
        self.image = self._images[0]
        self.rect = self.image.get_rect(midtop = [400, 440])

    def update(self, time):
        # Switch displayed scores every 4 seconds.
        time = time % (4000 * len(self._images))
        self.image = self._images[int(time / 4000)]


# A menu; a bunch of boxes displayed on carts, with callbacks
class Menu(object):
    def __init__(self, items):
        platforms = [Platform([400 + 260 * i, 390], s[0], i)
                     for i, s in enumerate(items)]
        credits = Credits()
        self._score = TopScores()
        sprites = RenderUpdates([credits, self._score])
        sprites.add(platforms)

        pos = 0
        em = EventManager()
        em.clear()
        screen = pygame.display.get_surface()

        em.get()
        quit = False

        screen.blit(Menu.bg, [0, 0])
        sprites.draw(screen)
        pygame.display.update()
        
        while not quit:
            for ev in em.get():
                if ev.type == PLAYER:
                    if ev.key == LEFT:
                        if pos != 0:
                            pos -= 1
                            for p in platforms: p.left()
                    elif ev.key == RIGHT:
                        if pos != len(items) - 1:
                            pos += 1
                            for p in platforms: p.right()
                    else:
                        try:
                            r = items[pos][1][ev.key](self, platforms[pos],
                                                      pos, ev.key)
                            if r:
                                # If the callback returns true, then we
                                # need to redraw the whole screen and
                                # reread our scores. (If it returns false,
                                # it probably means it just modified some
                                # cart text).
                                self._score.read_scores()
                                screen.blit(Menu.bg, [0, 0])
                                pygame.display.update()
                        except KeyError: pass

                elif ev.type == QUIT: quit = True

            sprites.update(pygame.time.get_ticks())
            pygame.display.update(sprites.draw(screen))
            sprites.clear(screen, Menu.bg)
            pygame.time.clock.tick(60)

# The most common kind of menu entry, a single callback function
# that we call when certain keys are pressed.
def entry(f, dirs = True):
    if dirs: return { UP: f, DOWN: f, ROT_CC: f, ROT_CW: f, CONFIRM: f }
    else:  return { ROT_CC: f, ROT_CW: f, CONFIRM: f }

# A "go back" callback function. By posting a QUIT event to the queue,
# the menu will exit and go to the previous screen.
break_menu = entry(lambda *args: pygame.event.post(pygame.event.Event(QUIT)),
                   False)

# A sprite displaying a small credits banner in the lower-left corner.
class Credits(Sprite, Singleton):
    def __init__(self, lines = CREDITS):
        Sprite.__init__(self)
        self._lines = [textfx.shadow(i, 20, [255, 255, 255]) for i in lines]
        self._idx = 0
        self._update = pygame.time.get_ticks() + 7000
        self._w = 220
        self._h = 19
        self.update(pygame.time.get_ticks())

    def update(self, t):
        self.image = pygame.Surface([self._w, self._h], SRCALPHA, 32)
        self.rect = self.image.get_rect(bottomleft = [0, 600])
        if self._update - t > 1000:
            txt = self._lines[self._idx]
            r = txt.get_rect(center = [self._w // 2, self._h // 2])
            self.image.blit(txt, r)
        elif t < self._update:
            p = (self._update - t) / 1000.0
            wy = int(self._h * p)
            idx1 = self._idx
            idx2 = (self._idx + 1) % len(self._lines)
            txt1 = self._lines[idx1]
            txt2 = self._lines[idx2]
            r1 = txt1.get_rect()
            r2 = txt2.get_rect()
            r2.centerx = r1.centerx = self._w // 2
            r2.top = wy
            r1.bottom = wy
            self.image.blit(txt1, r1)
            self.image.blit(txt2, r2)
        else:
            self._idx = (self._idx + 1) % len(self._lines)
            self._update = t + 4000
            txt = self._lines[self._idx]
            r = txt.get_rect(center = [self._w // 2, self._h // 2])
            self.image.blit(txt, r)



# unlocker.py -- display information about unlocked game features
# Copyright 2004 Joe Wreschnig <piman@sacredchao.net>
# Released under the terms of the GNU GPL v2.
__revision__ = "$Id: unlocker.py 286 2004-09-04 03:51:59Z piman $"

import pygame

import config
import textfx
import wipes

from characters import Character
from events import EventManager

from constants import *

# Display help text for unlocked features. This includes a lot of
# verbatim text we should move to a data file somewhere.
def init(*args):
    em = EventManager()
    disp = pygame.Surface([800, 600])

    disp.fill([0, 0, 0])
    font = textfx.WrapFont(30, 760)
    bg = Character.default.border([780, 100]).convert()

    bg.set_alpha(256)
    disp.blit(bg, [0, 0])
    t = font.render(
        "Welcome to the unlock screen! As you do more in this "
        "game, new features will sometimes appear. By playing the "
        "game at least once, you've managed to unlock this unlock "
        "screen, for example. Keep practicing, and more things will "
        "appear...")

    disp.blit(t, t.get_rect(midleft = [20, 60]))

    if config.getboolean("unlock", "single"):
        bg.set_alpha(256)
        disp.blit(bg, [0, 120])
        t = font.render(
            "Congratulations! By scoring over 20,000 points in a 2/3 versus "
            "match, you've unlocked single player mode. It's a lot "
            "harder, so you should practice your crystal building and "
            "combo setups.")
        disp.blit(t, t.get_rect(midleft = [20, 180]))
    else:
        bg.set_alpha(127)
        disp.blit(bg, [0, 120])

    if config.getboolean("unlock", "unixbros"):
        bg.set_alpha(256)
        disp.blit(bg, [0, 240])
        t = font.render(
            "By breaking a crystal of over 16 gems, you've unlocked two more "
            "characters - The Yuniks brothers. Both encourage a very "
            "unique style of attacks. Go check them out.")
        disp.blit(t, t.get_rect(midleft = [20, 300]))
    else:
        bg.set_alpha(127)
        disp.blit(bg, [0, 240])

    if config.getboolean("unlock", "combat"):
        bg.set_alpha(256)
        disp.blit(bg, [0, 360])
        t = font.render(
            "By getting at least a 4 chain, you've unlocked combat "
            "mode. If this is on, special attack rocks might appear in "
            "versus mode, that have strange effects on the other "
            "player. Enable it in the setup menu.")
        disp.blit(t, t.get_rect(midleft = [20, 420]))
    else:
        bg.set_alpha(127)
        disp.blit(bg, [0, 360])

    if config.getboolean("unlock", "cpuversus"):
        bg.set_alpha(256)
        disp.blit(bg, [0, 480])
        t = font.render(
            "You've made it at least halfway through arcade mode! "
            "You can now practice against any dwarf, on any "
            "of the AI levels.")
        disp.blit(t, t.get_rect(midleft = [20, 540]))
    else:
        bg.set_alpha(127)
        disp.blit(bg, [0, 480])

    wipes.wipe_in(disp)
    pygame.display.update()
    quit = False
    while not quit:
        for ev in em.wait():
             quit = (quit or
                     ev.type == QUIT or
                     (ev.type == PLAYER and
                      ev.key in [ROT_CC, ROT_CW, CONFIRM]))

    return True
